

class Renderer(object):

    def __init__(self):
        pass

    def plot(self, old_state, action, new_state, reward, done):
        pass

    def render(self, mode=None, min_time_step=1.0):
        pass



